from __future__ import annotations

import atexit
import os
import socket
import threading
from typing import Optional, Union

from dask.base import tokenize

import coiled

_clients = {}


def _shutdown(client):
    client.shutdown()


def run(
    *,
    software: Optional[str] = None,
    container: Optional[str] = None,
    vm_type: Optional[list[str]] = None,
    cpu: Optional[Union[int, list[int]]] = None,
    memory: Optional[Union[str, list[str]]] = None,
    gpu: Optional[bool] = None,
    account: Optional[str] = None,
    region: Optional[str] = None,
    arm: Optional[bool] = None,
):
    """
    Run a function on cloud infrastructure and return the result back locally.

    See the :class:`coiled.Cluster` docstring for parameter descriptions.

    .. warning::

        ``@coiled.run`` is an experimental feature and is subject to breaking changes.

    Cloud resources will be reused when running multiple decorated functions that
    have consistent hardware and software specifications.
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            # Construct `name` so that we reuse clusters when the same
            # cluster specification is given
            token = tokenize(
                (os.getpid(), threading.get_ident(), socket.getfqdn()),  # account for local parallelization
                software,
                container,
                vm_type,
                cpu,
                memory,
                gpu,
                account,
                region,
                arm,
            )
            name = f"run-{token[:8]}"

            environ = None
            if container and "rapidsai" in container:
                environ = {"DISABLE_JUPYTER": "true"}  # needed for "stable" RAPIDS image

            # Check to see if we have an existing cluster first
            global _clients
            try:
                client = _clients[name]
            except KeyError:
                cluster = coiled.Cluster(
                    name=name,
                    account=account,
                    n_workers=0,
                    scheduler_cpu=cpu,
                    scheduler_memory=memory,
                    software=software,
                    container=container,
                    scheduler_options={"idle_timeout": "24 hours"},
                    scheduler_vm_types=vm_type,
                    worker_vm_types=vm_type,
                    allow_ssh=True,
                    environ=environ,
                    scheduler_gpu=gpu,
                    region=region,
                    arm=arm,
                    tags={"coiled-cluster-type": "run/python"},
                )
                client = _clients[name] = cluster.get_client()
                atexit.register(_shutdown, client)

            return client.run_on_scheduler(func, *args, **kwargs)

        return wrapper

    return decorator
